// Ambrose Sterr 01/31/2006
// Finite State Machine Simulator - Simulation Handler
// For Formal Languages and the Theory of Computation @ Marlboro College
// Version 0.2
// Changes from Version 0.1:
// -Fixed a bug that carried the final state of one run into the next.
public class machine
{
	String[] alphabet, states;
	int numStates, start, alphabetSize;
	boolean[] accepts, curPos, nextPos, rules[][];
// The machine is created with its states, its start state, its accept states, and its alphabet.  Transitions are added later.
// The data is fed to the machine as string arrays.  It uses the array indexes to interpret the data and to handle execution,
// and only uses the string arrays to look up the indexes of states and characters.
// Position is stored as a boolean array of which states are active.
	public machine(String[] s, String sState, String[] accpts, String[] alph)
	{
		alphabet = alph;
		alphabetSize = alphabet.length;
		states = s;
		numStates = states.length;
		start = -1;
// Indexing the start state.
		for (int i = 0; (start == -1) && (i <= numStates); i++)
		{
			if (i == numStates)
			{
				System.out.println("Start state "+sState+" does not exist.\nEnding...");
				System.exit(0);
			}
			if (states[i].equals(sState))
				start = i;
		}
// Indexing the accept states.
		accepts = new boolean[numStates];
		for (int i = 0; i < numStates; i++)
		{
			accepts[i] = false;
			for (int j = 0; !(accepts[i]) && (j < accpts.length); j++)
				if (accpts[j].equals(states[i]))
					accepts[i] = true;
		}
// The transition array is initialized.
		rules = new boolean[numStates][alphabetSize+1][numStates];
		for (int i = 0; i < numStates; i++)
			for (int j = 0; j < (alphabetSize + 1); j++)
				for (int k = 0; k < numStates; k++)
				{
					rules[i][j][k] = false;
				}
	}
// This method "teaches" transitions to the machine.
	public void makeRule(String fromState, String character, String toState)
	{
		int fState = -1;
		int tState = -1;
		int charac = -1;
// Finding the indexes of the states involved.
		for (int i = 0; ((fState == -1) || (tState == -1)) && (i < numStates); i++)
		{
			if ((fState == -1) && (fromState.equals(states[i])))
				fState = i;
			if ((tState == -1) && (toState.equals(states[i])))
				tState = i;
		}
// Interpreting ~ as epsilon.
		if (character.equals("~"))
			charac = 0;
// Finding the indexes of the character.
		for (int i = 0; (charac == -1) && (i < alphabetSize); i++)
			if (character.equals(alphabet[i]))
				charac = i + 1;
// The machine will error on transitions it can't index, but will not halt.
		if ((fState == -1) || (tState == -1) || (charac == -1))
			System.out.println("Transition "+fromState+" x "+character+" -> "+toState+" doesn't make any sense.");
		else
			rules[fState][charac][tState] = true;
	}
// The execution method reads each input in the string, calculates the new position array, moves to it, and the grabs the next input.
	public boolean run(String[] inputs)
	{
// The current position is initialized.
		curPos = new boolean[numStates];
		nextPos = new boolean[numStates];
		for (int i = 0; i < numStates; i++)
			curPos[i] = false;
		curPos[start] = true;
// It checks for epsilons before reading any of the inputs.
		this.epsilon();
		for (int i = 0; i < inputs.length; i++)
		{
// Finding the index of the input character.
			int input = -1;
			for (int j = 0; (input == -1) && (j <= alphabetSize); j++)
			{
				if (j == alphabetSize)
				{
					System.out.println("Input "+inputs[i]+" is not in alphabet.");
					return false;
				}
				if (inputs[i].equals(alphabet[j]))
				{
// Input index 0 refers to epsilon.
					input = j+1;
				}
			}
// Clearing the next position array. 
			for (int j = 0; j < numStates; j++)
			{
				nextPos[j] = false;
			}
// Calculating which states the machine ends up in after this input.
			for (int j = 0; j < numStates; j++)
				if (curPos[j])
					for (int k = 0; k < numStates; k++)
						if (rules[j][input][k])
							nextPos[k] = true;
// Changing the current position to what has been calculated.
			for (int j = 0; j < numStates; j++)
				if (nextPos[j])
					curPos[j] = true;
				else
					curPos[j] = false;
// It checks for epsilon shifts between each input.
			this.epsilon();
		}
// Checking to see if the machine is in any accept states at the end.
		boolean accepted = false;
		for (int i = 0; !(accepted) && (i < numStates); i++)
			if (curPos[i] && accepts[i])
				accepted = true;
// And handing back the result.
		return accepted;
	}
// This recursive method handles epsilons.
	private void epsilon()
	{
		boolean shifted = false;
// If there is an epsilon transition from a state the machine is in to one it isn't, it enters that state.
		for (int i = 0; i < numStates; i++)
			if (curPos[i])
				for (int j = 0; j < numStates; j++)
					if (!(curPos[j]) && rules[i][0][j])
					{
						curPos[j] = true;
						shifted = true;
					}
// If the machine has changed positions from an epsilon, it checks again to see if it more epsilons are available.
		if (shifted)
			this.epsilon();
	}
// This toString prints a table of true/falses for the transitions from each state, with each column signifying a destination and each row signifying an input.
// I use it for debugging.
	public String toString()
	{
		String s = "";
		for (int i = 0; i < numStates; i++)
		{
			s = s + "\n";
			for (int j = 0; j < (alphabetSize + 1); j++)
			{
				for (int k = 0; k < numStates; k++)
					s = s +rules[i][j][k] + " ";
				s = s + "\n";
			}
		}
		return s;
	}
}